@@ -1,6 +1,6 @@
 /*
  * Copyright 2010 Jeff Garzik
- * Copyright 2012-2014 pooler
+ * Copyright 2012-2017 pooler
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -127,7 +127,6 @@ static int opt_retries = -1;
 static int opt_fail_pause = 30;
 int opt_timeout = 0;
 static int opt_scantime = 5;
-static const bool opt_time = true;
 static enum algos opt_algo = ALGO_SCRYPT;
 static int opt_scrypt_n = 1024;
 static int opt_n_threads;
@@ -136,7 +135,7 @@ static char *rpc_url;
 static char *rpc_userpass;
 static char *rpc_user, *rpc_pass;
 static int pk_script_size;
-static unsigned char pk_script[25];
+static unsigned char pk_script[42];
 static char coinbase_sig[101] = "";
 char *opt_cert;
 char *opt_proxy;
@@ -357,11 +356,22 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 	unsigned char (*merkle_tree)[32] = NULL;
 	bool coinbase_append = false;
 	bool submit_coinbase = false;
-	bool version_force = false;
-	bool version_reduce = false;
+	bool segwit = false;
 	json_t *tmp, *txa;
 	bool rc = false;
 
+	tmp = json_object_get(val, "rules");
+	if (tmp && json_is_array(tmp)) {
+		n = json_array_size(tmp);
+		for (i = 0; i < n; i++) {
+			const char *s = json_string_value(json_array_get(tmp, i));
+			if (!s)
+				continue;
+			if (!strcmp(s, "segwit") || !strcmp(s, "!segwit"))
+				segwit = true;
+		}
+	}
+
 	tmp = json_object_get(val, "mutable");
 	if (tmp && json_is_array(tmp)) {
 		n = json_array_size(tmp);
@@ -373,10 +383,6 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 				coinbase_append = true;
 			else if (!strcmp(s, "submit/coinbase"))
 				submit_coinbase = true;
-			else if (!strcmp(s, "version/force"))
-				version_force = true;
-			else if (!strcmp(s, "version/reduce"))
-				version_reduce = true;
 		}
 	}
 
@@ -393,14 +399,6 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 		goto out;
 	}
 	version = json_integer_value(tmp);
-	if (version > 2) {
-		if (version_reduce) {
-			version = 2;
-		} else if (!version_force) {
-			applog(LOG_ERR, "Unrecognized block version: %u", version);
-			goto out;
-		}
-	}
 
 	if (unlikely(!jobj_binary(val, "previousblockhash", prevhash, sizeof(prevhash)))) {
 		applog(LOG_ERR, "JSON invalid previousblockhash");
@@ -470,19 +468,56 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 		le32enc((uint32_t *)(cbtx+37), 0xffffffff); /* prev txout index */
 		cbtx_size = 43;
 		/* BIP 34: height in coinbase */
-		for (n = work->height; n; n >>= 8)
+		for (n = work->height; n; n >>= 8) {
 			cbtx[cbtx_size++] = n & 0xff;
+			if (n < 0x100 && n >= 0x80)
+				cbtx[cbtx_size++] = 0;
+		}
 		cbtx[42] = cbtx_size - 43;
 		cbtx[41] = cbtx_size - 42; /* scriptsig length */
 		le32enc((uint32_t *)(cbtx+cbtx_size), 0xffffffff); /* sequence */
 		cbtx_size += 4;
-		cbtx[cbtx_size++] = 1; /* out-counter */
+		cbtx[cbtx_size++] = segwit ? 2 : 1; /* out-counter */
 		le32enc((uint32_t *)(cbtx+cbtx_size), (uint32_t)cbvalue); /* value */
 		le32enc((uint32_t *)(cbtx+cbtx_size+4), cbvalue >> 32);
 		cbtx_size += 8;
 		cbtx[cbtx_size++] = pk_script_size; /* txout-script length */
 		memcpy(cbtx+cbtx_size, pk_script, pk_script_size);
 		cbtx_size += pk_script_size;
+		if (segwit) {
+			unsigned char (*wtree)[32] = calloc(tx_count + 2, 32);
+			memset(cbtx+cbtx_size, 0, 8); /* value */
+			cbtx_size += 8;
+			cbtx[cbtx_size++] = 38; /* txout-script length */
+			cbtx[cbtx_size++] = 0x6a; /* txout-script */
+			cbtx[cbtx_size++] = 0x24;
+			cbtx[cbtx_size++] = 0xaa;
+			cbtx[cbtx_size++] = 0x21;
+			cbtx[cbtx_size++] = 0xa9;
+			cbtx[cbtx_size++] = 0xed;
+			for (i = 0; i < tx_count; i++) {
+				const json_t *tx = json_array_get(txa, i);
+				const json_t *hash = json_object_get(tx, "hash");
+				if (!hash || !hex2bin(wtree[1+i], json_string_value(hash), 32)) {
+					applog(LOG_ERR, "JSON invalid transaction hash");
+					free(wtree);
+					goto out;
+				}
+				memrev(wtree[1+i], 32);
+			}
+			n = tx_count + 1;
+			while (n > 1) {
+				if (n % 2)
+					memcpy(wtree[n], wtree[n-1], 32);
+				n = (n + 1) / 2;
+				for (i = 0; i < n; i++)
+					sha256d(wtree[i], wtree[2*i], 64);
+			}
+			memset(wtree[1], 0, 32);  /* witness reserved value = 0 */
+			sha256d(cbtx+cbtx_size, wtree[0], 64);
+			cbtx_size += 32;
+			free(wtree);
+		}
 		le32enc((uint32_t *)(cbtx+cbtx_size), 0); /* lock time */
 		cbtx_size += 4;
 		coinbase_append = true;
@@ -545,13 +580,23 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 		tmp = json_array_get(txa, i);
 		const char *tx_hex = json_string_value(json_object_get(tmp, "data"));
 		const int tx_size = tx_hex ? strlen(tx_hex) / 2 : 0;
-		unsigned char *tx = malloc(tx_size);
-		if (!tx_hex || !hex2bin(tx, tx_hex, tx_size)) {
-			applog(LOG_ERR, "JSON invalid transactions");
+		if (segwit) {
+			const char *txid = json_string_value(json_object_get(tmp, "txid"));
+			if (!txid || !hex2bin(merkle_tree[1 + i], txid, 32)) {
+				applog(LOG_ERR, "JSON invalid transaction txid");
+				goto out;
+			}
+			memrev(merkle_tree[1 + i], 32);
+		} else {
+			unsigned char *tx = malloc(tx_size);
+			if (!tx_hex || !hex2bin(tx, tx_hex, tx_size)) {
+				applog(LOG_ERR, "JSON invalid transactions");
+				free(tx);
+				goto out;
+			}
+			sha256d(merkle_tree[1 + i], tx, tx_size);
 			free(tx);
-			goto out;
 		}
-		sha256d(merkle_tree[1 + i], tx, tx_size);
 		if (!submit_coinbase)
 			strcat(work->txs, tx_hex);
 	}
@@ -602,7 +647,7 @@ static bool gbt_work_decode(const json_t *val, struct work *work)
 		if (!have_longpoll) {
 			char *lp_uri;
 			tmp = json_object_get(val, "longpolluri");
-			lp_uri = json_is_string(tmp) ? strdup(json_string_value(tmp)) : rpc_url;
+			lp_uri = strdup(json_is_string(tmp) ? json_string_value(tmp) : rpc_url);
 			have_longpoll = true;
 			tq_push(thr_info[longpoll_thr_id].q, lp_uri);
 		}
@@ -658,19 +703,22 @@ static bool submit_upstream_work(CURL *curl, struct work *work)
 
 	if (have_stratum) {
 		uint32_t ntime, nonce;
-		char ntimestr[9], noncestr[9], *xnonce2str;
+		char ntimestr[9], noncestr[9], *xnonce2str, *req;
 
 		le32enc(&ntime, work->data[17]);
 		le32enc(&nonce, work->data[19]);
 		bin2hex(ntimestr, (const unsigned char *)(&ntime), 4);
 		bin2hex(noncestr, (const unsigned char *)(&nonce), 4);
 		xnonce2str = abin2hex(work->xnonce2, work->xnonce2_len);
-		sprintf(s,
+		req = malloc(256 + strlen(rpc_user) + strlen(work->job_id) + 2 * work->xnonce2_len);
+		sprintf(req,
 			"{\"method\": \"mining.submit\", \"params\": [\"%s\", \"%s\", \"%s\", \"%s\", \"%s\"], \"id\":4}",
 			rpc_user, work->job_id, xnonce2str, ntimestr, noncestr);
 		free(xnonce2str);
 
-		if (unlikely(!stratum_send_line(&stratum, s))) {
+		rc = stratum_send_line(&stratum, req);
+		free(req);
+		if (unlikely(!rc)) {
 			applog(LOG_ERR, "submit_upstream_work stratum_send_line failed");
 			goto out;
 		}
@@ -758,13 +806,14 @@ static const char *getwork_req =
 	"{\"method\": \"getwork\", \"params\": [], \"id\":0}\r\n";
 
 #define GBT_CAPABILITIES "[\"coinbasetxn\", \"coinbasevalue\", \"longpoll\", \"workid\"]"
+#define GBT_RULES "[\"segwit\"]"
 
 static const char *gbt_req =
 	"{\"method\": \"getblocktemplate\", \"params\": [{\"capabilities\": "
-	GBT_CAPABILITIES "}], \"id\":0}\r\n";
+	GBT_CAPABILITIES ", \"rules\": " GBT_RULES "}], \"id\":0}\r\n";
 static const char *gbt_lp_req =
 	"{\"method\": \"getblocktemplate\", \"params\": [{\"capabilities\": "
-	GBT_CAPABILITIES ", \"longpollid\": \"%s\"}], \"id\":0}\r\n";
+	GBT_CAPABILITIES ", \"rules\": " GBT_RULES ", \"longpollid\": \"%s\"}], \"id\":0}\r\n";
 
 static bool get_upstream_work(CURL *curl, struct work *work)
 {
@@ -1110,6 +1159,7 @@ static void *miner_thread(void *userdata)
 			if (!have_stratum &&
 			    (time(NULL) - g_work_time >= min_scantime ||
 			     work.data[19] >= end_nonce)) {
+				work_free(&g_work);
 				if (unlikely(!get_work(mythr, &g_work))) {
 					applog(LOG_ERR, "work retrieval failed, exiting "
 						"mining thread %d", mythr->id);
@@ -1282,6 +1332,7 @@ start:
 			soval = json_object_get(res, "submitold");
 			submit_old = soval ? json_is_true(soval) : false;
 			pthread_mutex_lock(&g_work_lock);
+			work_free(&g_work);
 			if (have_gbt)
 				rc = gbt_work_decode(res, &g_work);
 			else
@@ -1422,6 +1473,7 @@ static void show_version_and_exit(void)
 #endif
 #if defined(USE_ASM) && defined(__x86_64__)
 		" x86_64"
+		" PHE"
 #endif
 #if defined(USE_ASM) && (defined(__i386__) || defined(__x86_64__))
 		" SSE2"
@@ -1450,6 +1502,12 @@ static void show_version_and_exit(void)
 #if defined(__ARM_NEON__)
 		" NEON"
 #endif
+#endif
+#if defined(USE_ASM) && (defined(__powerpc__) || defined(__ppc__) || defined(__PPC__))
+		" PowerPC"
+#if defined(__ALTIVEC__)
+		" AltiVec"
+#endif
 #endif
 		"\n");
 
@@ -1605,7 +1663,8 @@ static void parse_arg(int key, char *arg, char *pname)
 		if (ap != arg) {
 			if (strncasecmp(arg, "http://", 7) &&
 			    strncasecmp(arg, "https://", 8) &&
-			    strncasecmp(arg, "stratum+tcp://", 14)) {
+			    strncasecmp(arg, "stratum+tcp://", 14) &&
+			    strncasecmp(arg, "stratum+tcps://", 15)) {
 				fprintf(stderr, "%s: unknown protocol -- '%s'\n",
 					pname, arg);
 				show_usage_and_exit(1);
@@ -1816,7 +1875,8 @@ int main(int argc, char *argv[])
 	pthread_mutex_init(&stratum.sock_lock, NULL);
 	pthread_mutex_init(&stratum.work_lock, NULL);
 
-	flags = !opt_benchmark && strncmp(rpc_url, "https:", 6)
+	flags = opt_benchmark || (strncasecmp(rpc_url, "https://", 8) &&
+	                          strncasecmp(rpc_url, "stratum+tcps://", 15))
 	      ? (CURL_GLOBAL_ALL & ~CURL_GLOBAL_SSL)
 	      : CURL_GLOBAL_ALL;
 	if (curl_global_init(flags)) {
