@@ -1,7 +1,8 @@
 /*
  * Copyright 2010 Jeff Garzik
  * Copyright 2012 Luke Dashjr
- * Copyright 2012-2014 pooler
+ * Copyright 2012-2020 pooler
+ * Copyright 2017 Pieter Wuille
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -507,6 +508,16 @@ err_out:
 	return NULL;
 }
 
+void memrev(unsigned char *p, size_t len)
+{
+	unsigned char c, *q;
+	for (q = p + len - 1; p < q; p++, q--) {
+		c = *p;
+		*p = *q;
+		*q = c;
+	}
+}
+
 void bin2hex(char *s, const unsigned char *p, size_t len)
 {
 	int i;
@@ -649,6 +660,136 @@ static int b58check(unsigned char *bin, size_t binsz, const char *b58)
 	return bin[0];
 }
 
+static uint32_t bech32_polymod_step(uint32_t pre) {
+	uint8_t b = pre >> 25;
+	return ((pre & 0x1FFFFFF) << 5) ^
+		(-((b >> 0) & 1) & 0x3b6a57b2UL) ^
+		(-((b >> 1) & 1) & 0x26508e6dUL) ^
+		(-((b >> 2) & 1) & 0x1ea119faUL) ^
+		(-((b >> 3) & 1) & 0x3d4233ddUL) ^
+		(-((b >> 4) & 1) & 0x2a1462b3UL);
+}
+
+static const int8_t bech32_charset_rev[128] = {
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+	15, -1, 10, 17, 21, 20, 26, 30,  7,  5, -1, -1, -1, -1, -1, -1,
+	-1, 29, -1, 24, 13, 25,  9,  8, 23, -1, 18, 22, 31, 27, 19, -1,
+	 1,  0,  3, 16, 11, 28, 12, 14,  6,  4,  2, -1, -1, -1, -1, -1,
+	-1, 29, -1, 24, 13, 25,  9,  8, 23, -1, 18, 22, 31, 27, 19, -1,
+	 1,  0,  3, 16, 11, 28, 12, 14,  6,  4,  2, -1, -1, -1, -1, -1
+};
+
+static bool bech32_decode(char *hrp, uint8_t *data, size_t *data_len, const char *input) {
+	uint32_t chk = 1;
+	size_t i;
+	size_t input_len = strlen(input);
+	size_t hrp_len;
+	int have_lower = 0, have_upper = 0;
+	if (input_len < 8 || input_len > 90) {
+		return false;
+	}
+	*data_len = 0;
+	while (*data_len < input_len && input[(input_len - 1) - *data_len] != '1') {
+		++(*data_len);
+	}
+	hrp_len = input_len - (1 + *data_len);
+	if (1 + *data_len >= input_len || *data_len < 6) {
+		return false;
+	}
+	*(data_len) -= 6;
+	for (i = 0; i < hrp_len; ++i) {
+		int ch = input[i];
+		if (ch < 33 || ch > 126) {
+			return false;
+		}
+		if (ch >= 'a' && ch <= 'z') {
+			have_lower = 1;
+		} else if (ch >= 'A' && ch <= 'Z') {
+			have_upper = 1;
+			ch = (ch - 'A') + 'a';
+		}
+		hrp[i] = ch;
+		chk = bech32_polymod_step(chk) ^ (ch >> 5);
+	}
+	hrp[i] = 0;
+	chk = bech32_polymod_step(chk);
+	for (i = 0; i < hrp_len; ++i) {
+		chk = bech32_polymod_step(chk) ^ (input[i] & 0x1f);
+	}
+	++i;
+	while (i < input_len) {
+		int v = (input[i] & 0x80) ? -1 : bech32_charset_rev[(int)input[i]];
+		if (input[i] >= 'a' && input[i] <= 'z') have_lower = 1;
+		if (input[i] >= 'A' && input[i] <= 'Z') have_upper = 1;
+		if (v == -1) {
+			return false;
+		}
+		chk = bech32_polymod_step(chk) ^ v;
+		if (i + 6 < input_len) {
+			data[i - (1 + hrp_len)] = v;
+		}
+		++i;
+	}
+	if (have_lower && have_upper) {
+		return false;
+	}
+	return chk == 1;
+}
+
+static bool convert_bits(uint8_t *out, size_t *outlen, int outbits, const uint8_t *in, size_t inlen, int inbits, int pad) {
+	uint32_t val = 0;
+	int bits = 0;
+	uint32_t maxv = (((uint32_t)1) << outbits) - 1;
+	while (inlen--) {
+		val = (val << inbits) | *(in++);
+		bits += inbits;
+		while (bits >= outbits) {
+			bits -= outbits;
+			out[(*outlen)++] = (val >> bits) & maxv;
+		}
+	}
+	if (pad) {
+		if (bits) {
+			out[(*outlen)++] = (val << (outbits - bits)) & maxv;
+		}
+	} else if (((val << (outbits - bits)) & maxv) || bits >= inbits) {
+		return false;
+	}
+	return true;
+}
+
+static bool segwit_addr_decode(int *witver, uint8_t *witdata, size_t *witdata_len, const char *addr) {
+	uint8_t data[84];
+	char hrp_actual[84];
+	size_t data_len;
+	if (!bech32_decode(hrp_actual, data, &data_len, addr)) return false;
+	if (data_len == 0 || data_len > 65) return false;
+	if (data[0] > 16) return false;
+	*witdata_len = 0;
+	if (!convert_bits(witdata, witdata_len, 8, data + 1, data_len - 1, 5, 0)) return false;
+	if (*witdata_len < 2 || *witdata_len > 40) return false;
+	if (data[0] == 0 && *witdata_len != 20 && *witdata_len != 32) return false;
+	*witver = data[0];
+	return true;
+}
+
+static size_t bech32_to_script(uint8_t *out, size_t outsz, const char *addr) {
+	uint8_t witprog[40];
+	size_t witprog_len;
+	int witver;
+
+	if (!segwit_addr_decode(&witver, witprog, &witprog_len, addr))
+		return 0;
+	if (outsz < witprog_len + 2)
+		return 0;
+	out[0] = witver ? (0x50 + witver) : 0;
+	out[1] = witprog_len;
+	memcpy(out + 2, witprog, witprog_len);
+	return witprog_len + 2;
+}
+
 size_t address_to_script(unsigned char *out, size_t outsz, const char *addr)
 {
 	unsigned char addrbin[25];
@@ -656,7 +797,7 @@ size_t address_to_script(unsigned char *out, size_t outsz, const char *addr)
 	size_t rv;
 
 	if (!b58dec(addrbin, sizeof(addrbin), addr))
-		return 0;
+		return bech32_to_script(out, outsz, addr);
 	addrver = b58check(addrbin, sizeof(addrbin), addr);
 	if (addrver < 0)
 		return 0;
@@ -770,7 +911,7 @@ void diff_to_target(uint32_t *target, double diff)
 #define socket_blocks() (errno == EAGAIN || errno == EWOULDBLOCK)
 #endif
 
-static bool send_line(curl_socket_t sock, char *s)
+static bool send_line(struct stratum_ctx *sctx, char *s)
 {
 	ssize_t len, sent = 0;
 	
@@ -783,12 +924,18 @@ static bool send_line(curl_socket_t sock, char *s)
 		fd_set wd;
 
 		FD_ZERO(&wd);
-		FD_SET(sock, &wd);
-		if (select(sock + 1, NULL, &wd, NULL, &timeout) < 1)
+		FD_SET(sctx->sock, &wd);
+		if (select(sctx->sock + 1, NULL, &wd, NULL, &timeout) < 1)
 			return false;
-		n = send(sock, s + sent, len, 0);
+#if LIBCURL_VERSION_NUM >= 0x071202
+		CURLcode rc = curl_easy_send(sctx->curl, s + sent, len, (size_t *)&n);
+		if (rc != CURLE_OK) {
+			if (rc != CURLE_AGAIN)
+#else
+		n = send(sctx->sock, s + sent, len, 0);
 		if (n < 0) {
 			if (!socket_blocks())
+#endif
 				return false;
 			n = 0;
 		}
@@ -807,7 +954,7 @@ bool stratum_send_line(struct stratum_ctx *sctx, char *s)
 		applog(LOG_DEBUG, "> %s", s);
 
 	pthread_mutex_lock(&sctx->sock_lock);
-	ret = send_line(sctx->sock, s);
+	ret = send_line(sctx, s);
 	pthread_mutex_unlock(&sctx->sock_lock);
 
 	return ret;
@@ -867,6 +1014,15 @@ char *stratum_recv_line(struct stratum_ctx *sctx)
 			ssize_t n;
 
 			memset(s, 0, RBUFSIZE);
+#if LIBCURL_VERSION_NUM >= 0x071202
+			CURLcode rc = curl_easy_recv(sctx->curl, s, RECVSIZE, (size_t *)&n);
+			if (rc == CURLE_OK && !n) {
+				ret = false;
+				break;
+			}
+			if (rc != CURLE_OK) {
+				if (rc != CURLE_AGAIN || !socket_full(sctx->sock, 1)) {
+#else
 			n = recv(sctx->sock, s, RECVSIZE, 0);
 			if (!n) {
 				ret = false;
@@ -874,6 +1030,7 @@ char *stratum_recv_line(struct stratum_ctx *sctx)
 			}
 			if (n < 0) {
 				if (!socket_blocks() || !socket_full(sctx->sock, 1)) {
+#endif
 					ret = false;
 					break;
 				}
@@ -907,7 +1064,7 @@ out:
 	return sret;
 }
 
-#if LIBCURL_VERSION_NUM >= 0x071101
+#if LIBCURL_VERSION_NUM >= 0x071101 && LIBCURL_VERSION_NUM < 0x072d00
 static curl_socket_t opensocket_grab_cb(void *clientp, curlsocktype purpose,
 	struct curl_sockaddr *addr)
 {
@@ -945,11 +1102,13 @@ bool stratum_connect(struct stratum_ctx *sctx, const char *url)
 	}
 	free(sctx->curl_url);
 	sctx->curl_url = malloc(strlen(url));
-	sprintf(sctx->curl_url, "http%s", strstr(url, "://"));
+	sprintf(sctx->curl_url, "http%s", url + 11);
 
 	if (opt_protocol)
 		curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
 	curl_easy_setopt(curl, CURLOPT_URL, sctx->curl_url);
+	if (opt_cert)
+		curl_easy_setopt(curl, CURLOPT_CAINFO, opt_cert);
 	curl_easy_setopt(curl, CURLOPT_FRESH_CONNECT, 1);
 	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 30);
 	curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, sctx->curl_err_str);
@@ -963,7 +1122,7 @@ bool stratum_connect(struct stratum_ctx *sctx, const char *url)
 #if LIBCURL_VERSION_NUM >= 0x070f06
 	curl_easy_setopt(curl, CURLOPT_SOCKOPTFUNCTION, sockopt_keepalive_cb);
 #endif
-#if LIBCURL_VERSION_NUM >= 0x071101
+#if LIBCURL_VERSION_NUM >= 0x071101 && LIBCURL_VERSION_NUM < 0x072d00
 	curl_easy_setopt(curl, CURLOPT_OPENSOCKETFUNCTION, opensocket_grab_cb);
 	curl_easy_setopt(curl, CURLOPT_OPENSOCKETDATA, &sctx->sock);
 #endif
@@ -977,7 +1136,9 @@ bool stratum_connect(struct stratum_ctx *sctx, const char *url)
 		return false;
 	}
 
-#if LIBCURL_VERSION_NUM < 0x071101
+#if LIBCURL_VERSION_NUM >= 0x072d00
+	curl_easy_getinfo(curl, CURLINFO_ACTIVESOCKET, &sctx->sock);
+#elif LIBCURL_VERSION_NUM < 0x071101
 	/* CURLINFO_LASTSOCKET is broken on Win64; only use it as a last resort */
 	curl_easy_getinfo(curl, CURLINFO_LASTSOCKET, (long *)&sctx->sock);
 #endif
@@ -1289,7 +1450,8 @@ static bool stratum_reconnect(struct stratum_ctx *sctx, json_t *params)
 		return false;
 
 	url = malloc(32 + strlen(host));
-	sprintf(url, "stratum+tcp://%s:%d", host, port);
+	strncpy(url, sctx->url, 15);
+	sprintf(strstr(url, "://") + 3, "%s:%d", host, port);
 
 	if (!opt_redirect) {
 		applog(LOG_INFO, "Ignoring request to reconnect to %s", url);
